==> g3dviewer-0.0.3.tar.gz: g3dviewer-0.0.3/src/gl.c <==
#include <stdlib.h>
#include <math.h>

#include <glib.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <g3dviewer_plugin.h>

#include "main.h"
#include "gl.h"
#include "trackball.h"

#if DEBUG > 1
#define TIMING
#endif

static int _initialized = 0;

#ifdef TIMING
static GTimer *timer = NULL;
static gulong avg_msec = 0;
#endif

int gl_normal_unify(GLfloat *nx, GLfloat *ny, GLfloat *nz);

void gl_init(void)
{
#if DEBUG > 1
  g_printerr("init OpenGL\n");
#endif

  GLfloat light0_pos[4] = { -50.0, 50.0, 0.0, 0.0 };
  GLfloat light0_col[4] = { 0.6, 0.6, 0.6, 1.0 };
  GLfloat light1_pos[4] = {  50.0, 50.0, 0.0, 0.0 };
  GLfloat light1_col[4] = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat ambient_lc[4] = { 0.25, 0.25, 0.25, 1.0 };

  glEnable(GL_CULL_FACE);

  glEnable(GL_ALPHA_TEST);
 
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);

  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);

#if 0
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_POLYGON_SMOOTH);
#endif

#if 0
  glEnable(GL_DITHER);
#endif
  glShadeModel(GL_SMOOTH);
  
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_lc);
  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_col);
  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_col);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHTING);

  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

#ifdef TIMING
  timer = g_timer_new();
#endif
}

void gl_draw(G3DModel *model)
{
  GSList *olist;
  int i, j;
  G3DMaterial *prev_material = NULL;
  G3DObject *object;
  GLfloat m[4][4];
#ifdef OLD_ALGO
  int prev_index_cnt = -1;
  GSList *flist;
  G3DFace *face;
  GLfloat prev_nx = 0;
  GLfloat prev_ny = 0;
  GLfloat prev_nz = 0;
  GLfloat ax,ay,az, bx,by,bz, nx,ny,nz;
#endif

  if(! _initialized)
  {
    gl_init();
    _initialized = 1;
  }

  /* draw scene... */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(model->zoom, model->aspect, 1, 100);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(model->bgcolor[0], model->bgcolor[1], model->bgcolor[2], 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  
  glLoadIdentity();
  glTranslatef(0, 0, -30);
  build_rotmatrix(m, model->quat);
  glMultMatrixf(&m[0][0]);

  g_return_if_fail(model != NULL);

#ifdef TIMING
  g_timer_start(timer);
#endif

  olist = model->objects;
  while(olist != NULL)
  {
#ifdef OLD_ALGO
    int fcnt = 0;
#endif
    object = (G3DObject *)olist->data;

    g_return_if_fail(object != NULL);
#if DEBUG > 3
    g_printerr("name: %s {", object->name);
#endif

/*****************************************************************************/
/* newer drawing algorithm, draws only triangles, fewer comparisons          */

#ifndef OLD_ALGO /* draw triangles from array */
    glBegin(GL_TRIANGLES);      
    for(i=0; i<object->_num_faces; i++)
    {
      if(prev_material != object->_materials[i])
      {
        GLfloat normspec[4] = { 0.0, 0.0, 0.0, 1.0 };

        g_return_if_fail(object->_materials[i] != NULL);

        prev_material = object->_materials[i];
        glColor4f(prev_material->r, prev_material->g, 
                  prev_material->b, prev_material->a);
        if(model->glflags & G3D_FLAG_GL_SPECULAR)
          glMaterialfv(GL_FRONT, GL_SPECULAR, prev_material->specular);
        else glMaterialfv(GL_FRONT, GL_SPECULAR, normspec);
        if(model->glflags & G3D_FLAG_GL_SHININESS)
          glMaterialf(GL_FRONT, GL_SHININESS, prev_material->shininess);
        else glMaterialf(GL_FRONT, GL_SHININESS, 0.0);
      }

      for(j=0; j<3; j++)
      {
        glNormal3f(object->_normals[(i*3+j)*3+0],
                   object->_normals[(i*3+j)*3+1],
                   object->_normals[(i*3+j)*3+2]);
        glVertex3f(object->vertices[object->_indices[i*3+j]*3+0],
                   object->vertices[object->_indices[i*3+j]*3+1],
                   object->vertices[object->_indices[i*3+j]*3+2]);
      }
      if(model->glflags & G3D_FLAG_GL_ALLTWOSIDE)
      {
        for(j=2; j>=0; j--)
        {
          if(object->_flags[i] & G3D_FLAG_FAC_NORMALS)
          {
            glNormal3f(object->_normals[(i*3+j)*3+0],
                       object->_normals[(i*3+j)*3+1],
                       object->_normals[(i*3+j)*3+2]);
          }
          else
          {
            glNormal3f(-object->_normals[(i*3+j)*3+0],
                       -object->_normals[(i*3+j)*3+1],
                       -object->_normals[(i*3+j)*3+2]);
          }
          glVertex3f(object->vertices[object->_indices[i*3+j]*3+0],
                     object->vertices[object->_indices[i*3+j]*3+1],
                     object->vertices[object->_indices[i*3+j]*3+2]);
        }
      }
    }
    glEnd();

/*****************************************************************************/
/* original drawing algorithm, _SHOULD_ be slower                            */

#else /* faces from GSList */

#define PX(i) (object->vertices[face->index_vertex[i]*3+0])
#define PY(i) (object->vertices[face->index_vertex[i]*3+1])
#define PZ(i) (object->vertices[face->index_vertex[i]*3+2])

    flist = object->faces;
    while(flist != NULL)
    { 
      face = (G3DFace*)flist->data;
           
      g_return_if_fail(face != NULL);
      fcnt++;
#if 0
      g_printerr(" %d\n", fcnt);
#endif
      if(face->num_vertices < 3) 
      { 
        flist = flist->next;
        continue;
      }

      ax = PX(1) - PX(0);
      ay = PY(1) - PY(0);
      az = PZ(1) - PZ(0);

      bx = PX(face->num_vertices - 1) - PX(0);
      by = PY(face->num_vertices - 1) - PY(0);
      bz = PZ(face->num_vertices - 1) - PZ(0);

      nx = ay * bz - az * by;
      ny = az * bx - ax * bz;
      nz = ax * by - ay * bx;

      if(gl_normal_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
      {
        flist = flist->next;
        continue;
      }

      if(prev_index_cnt != face->num_vertices || prev_index_cnt > 4)
      {
        if(prev_index_cnt > 0) glEnd();
        prev_index_cnt = face->num_vertices;
        switch(face->num_vertices)
        {
          case 3:  glBegin(GL_TRIANGLES); break;
          case 4:  glBegin(GL_QUADS); break;
          default: glBegin(GL_POLYGON);
        }
      }

      if(prev_material != face->material)
      {
        GLfloat normspec[4] = { 0.0, 0.0, 0.0, 1.0 };

        g_return_if_fail(face->material != NULL);
        prev_material = face->material;
        glColor4f(face->material->r, face->material->g, 
                  face->material->b, face->material->a);
        if(model->glflags & G3D_FLAG_GL_SPECULAR)
          glMaterialfv(GL_FRONT, GL_SPECULAR, face->material->specular);
        else glMaterialfv(GL_FRONT, GL_SPECULAR, normspec);
        if(model->glflags & G3D_FLAG_GL_SHININESS)
          glMaterialf(GL_FRONT, GL_SHININESS, face->material->shininess);
        else glMaterialf(GL_FRONT, GL_SHININESS, 0.0);
      }
#if DEBUG > 3
      g_printerr(" mat");
#endif
      if(nx != prev_nx || ny != prev_ny || nz != prev_nz)
      {
        prev_nx = nx;
        prev_ny = ny;
        prev_nz = nz;
        glNormal3f(nx, ny, nz);
      }
#if DEBUG > 3
      g_printerr(" nrm");
#endif
      for(i=0; i<face->num_vertices; i++)
      {
        if(face->flags & G3D_FLAG_FAC_NORMALS)
        {
          nx = face->normals[i*3+0];
          ny = face->normals[i*3+1];
          nz = face->normals[i*3+2];

          if(gl_normal_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
          {
            flist = flist->next;
            continue;
          }
          
          if(nx != prev_nx || ny != prev_ny || nz != prev_nz)
          {
            prev_nx = nx;
            prev_ny = ny;
            prev_nz = nz;
            glNormal3f(nx, ny, nz);
          } 
        }
        glVertex3f(PX(i), PY(i), PZ(i));
      }

      if((face->material->flags & G3D_FLAG_MAT_TWOSIDE) || 
         (model->glflags & G3D_FLAG_GL_ALLTWOSIDE))
      {
        if(!(face->flags & G3D_FLAG_FAC_NORMALS))
        {
          prev_nx = nx = -nx;
          prev_ny = ny = -ny;
          prev_nz = nz = -nz;
          glNormal3f(nx, ny, nz);
        }
        for(i=face->num_vertices-1; i>=0; i--) 
        {
          if(face->flags & G3D_FLAG_FAC_NORMALS)
          {
            nx = face->normals[i*3+0];
            ny = face->normals[i*3+1];
            nz = face->normals[i*3+2];

            if(gl_normal_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
            {
              flist = flist->next;
              continue;
            }
          
            if(nx != prev_nx || ny != prev_ny || nz != prev_nz)
            {
              prev_nx = nx;
              prev_ny = ny;
              prev_nz = nz;
              glNormal3f(nx, ny, nz);
            } 
          } 
          glVertex3f(PX(i), PY(i), PZ(i));
        }
      }
#if DEBUG > 3
      g_printerr(" fac");
#endif
      flist = flist->next;
    }
#endif /* draw faces from GSList instead of triangles */

#ifdef TIMING /* get time to draw one frame to compare algorithms */
    g_timer_stop(timer);

    if(avg_msec == 0)
    {
      gulong msec;
      gdouble sec;

      sec = g_timer_elapsed(timer, &msec);
      avg_msec = (gulong)sec * 1000000 + msec;
    }
    else
    {
      gulong msec, add;
      gdouble sec;

      sec = g_timer_elapsed(timer, &msec);
      add = (gulong)sec * 1000000 + msec;
      avg_msec = (avg_msec + add) / 2;
    }

    g_printerr("average time to render frame: %lu s\n", avg_msec);
#endif

#if DEBUG > 3
    g_printerr(" fcs");
    g_printerr(" }\n");
#endif

#ifdef OLD_ALGO
    if(prev_index_cnt > 0) glEnd();
    prev_index_cnt = -1;
    prev_nx = 0;
    prev_ny = 0;
    prev_nz = 0;
#endif

    olist = olist->next;
  } /* while olist != NULL */
#if DEBUG > 3
  g_printerr("gl.c: drawn...\n");
#endif
}

/*****************************************************************************/

int gl_normal_unify(GLfloat *nx, GLfloat *ny, GLfloat *nz)
{
  GLfloat r;

  r = sqrt(*nx * *nx + *ny * *ny + *nz * *nz);
  if(r < 0.000001) return EXIT_FAILURE;

  *nx /= r;
  *ny /= r;
  *nz /= r;
  return EXIT_SUCCESS;
}

==> g3dviewer-0.1.1.tar.gz: g3dviewer-0.1.1/src/gl.c <==
#include <stdlib.h>
#include <math.h>

#include <glib.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include <g3dviewer_plugin.h>

#include "main.h"
#include "gl.h"
#include "image.h"
#include "trackball.h"

#if DEBUG > 1
#define TIMING
#endif

static int _initialized = 0;

#ifdef TIMING
static GTimer *timer = NULL;
static gulong avg_msec = 0;
#endif

int gl_normal_unify(GLfloat *nx, GLfloat *ny, GLfloat *nz);

GLuint evil;

void gl_init(void)
{
#if DEBUG > 1
  g_printerr("init OpenGL\n");
#endif

  GLfloat light0_pos[4] = { -50.0, 50.0, 0.0, 0.0 };
  GLfloat light0_col[4] = { 0.6, 0.6, 0.6, 1.0 };
  GLfloat light1_pos[4] = {  50.0, 50.0, 0.0, 0.0 };
  GLfloat light1_col[4] = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat ambient_lc[4] = { 0.25, 0.25, 0.25, 1.0 };

  glEnable(GL_CULL_FACE);

  glEnable(GL_ALPHA_TEST);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);

  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);

#if 0
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_POLYGON_SMOOTH);
#endif

#if 0
  glEnable(GL_DITHER);
#endif
  glShadeModel(GL_SMOOTH);

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_lc);
  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_col);
  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_col);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHTING);

  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  /* texture stuff */
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );


#ifdef TIMING
  timer = g_timer_new();
#endif
}

void gl_prepare_texture(G3DImage *texture)
{
  guint w = 1, h = 1;

  while(w < texture->width) w *= 2;
  while(h < texture->height) h *= 2;

  if(w < h) w = h;
  else if(h < w) h = w;

  if((h == texture->height) && (w == texture->width)) return;

#if DEBUG > 0
  g_printerr("resizing texture to %ux%u\n", w, h);
#endif

  image_expand(texture, w, h);

#if DEBUG > 5
  g_printerr("resizing done...\n");
#endif
}

void gl_update_material(G3DModel *model, G3DMaterial *material)
{
  GLfloat normspec[4] = { 0.0, 0.0, 0.0, 1.0 };

  g_return_if_fail(material != NULL);

  glColor4f(material->r, material->g,
            material->b, material->a);
  if(model->glflags & G3D_FLAG_GL_SPECULAR)
    glMaterialfv(GL_FRONT, GL_SPECULAR, material->specular);
  else glMaterialfv(GL_FRONT, GL_SPECULAR, normspec);

  if(model->glflags & G3D_FLAG_GL_SHININESS)
    glMaterialf(GL_FRONT, GL_SHININESS, material->shininess);
  else glMaterialf(GL_FRONT, GL_SHININESS, 0.0);

  if(material->flags & G3D_FLAG_FAC_TEXMAP)
  {
    g_return_if_fail(material->texture != NULL);
    if(material->texture->depth == 32)
    {
      GLuint tex[1];
#if DEBUG > 3
      g_printerr("texturing image...\n");
#endif
      glGenTextures(1, tex);
      evil = *tex;

      gl_prepare_texture(material->texture);

      glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
#if 0
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
#endif
      glBindTexture( GL_TEXTURE_2D, *tex);
      glTexImage2D(GL_TEXTURE_2D, 1,
                   GL_RGBA,                       /* internal format */
                   material->texture->width,
                   material->texture->height,
                   0,                             /* border */
                   GL_RGBA,                       /* format of data */
                   GL_UNSIGNED_BYTE,              /* data type */
                   material->texture->pixeldata);
    }
    else
    {
      g_printerr("WARNING: unsupported color depth: %d bit\n",
                 material->texture->depth);
    }
  }
}

void gl_draw(G3DModel *model)
{
  GSList *olist;
  int i, j;
  G3DMaterial *prev_material = NULL;
  G3DObject *object;
  GLfloat m[4][4];
#ifdef OLD_ALGO
  int prev_index_cnt = -1;
  GSList *flist;
  G3DFace *face;
  GLfloat prev_nx = 0;
  GLfloat prev_ny = 0;
  GLfloat prev_nz = 0;
  GLfloat ax,ay,az, bx,by,bz, nx,ny,nz;
#endif

  if(! _initialized)
  {
    gl_init();
    _initialized = 1;
  }

  /* draw scene... */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(model->zoom, model->aspect, 1, 100);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(model->bgcolor[0], model->bgcolor[1], model->bgcolor[2], 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();
  glTranslatef(0, 0, -30);
  build_rotmatrix(m, model->quat);
  glMultMatrixf(&m[0][0]);

  g_return_if_fail(model != NULL);

#ifdef TIMING
  g_timer_start(timer);
#endif

  olist = model->objects;
  while(olist != NULL)
  {
#ifdef OLD_ALGO
    int fcnt = 0;
#endif
    object = (G3DObject *)olist->data;

    g_return_if_fail(object != NULL);
#if DEBUG > 3
    g_printerr("name: %s {", object->name);
#endif

/*****************************************************************************/
/* newer drawing algorithm, draws only triangles, fewer comparisons          */

#if DEBUG > 2
    g_printerr("new object\n");
#endif

#ifndef OLD_ALGO /* draw triangles from array */
    glBegin(GL_TRIANGLES);

    for(i=0; i<object->_num_faces; i++)
    {
      if(prev_material != object->_materials[i])
      {
        gl_update_material(model, object->_materials[i]);
        prev_material = object->_materials[i];
      }

      for(j=0; j<3; j++)
      {
        if(prev_material->flags & G3D_FLAG_FAC_TEXMAP)
        {
          glEnable(GL_TEXTURE_2D);

          if(object->tex_nverts > object->_indices[i*3+j])
          {
#if DEBUG > 5
            g_printerr("tex: %f x %f\n",
              object->tex_vertices[object->_indices[i*3+j]*2+0],
              object->tex_vertices[object->_indices[i*3+j]*2+1]);
#endif
            glBindTexture(GL_TEXTURE_2D, evil);
            glTexCoord2f(
              object->tex_vertices[object->_indices[i*3+j]*2+0],
              object->tex_vertices[object->_indices[i*3+j]*2+1]);
          }
          glDisable(GL_TEXTURE_2D);
        }
        glNormal3f(object->_normals[(i*3+j)*3+0],
                   object->_normals[(i*3+j)*3+1],
                   object->_normals[(i*3+j)*3+2]);
        glVertex3f(object->vertices[object->_indices[i*3+j]*3+0],
                   object->vertices[object->_indices[i*3+j]*3+1],
                   object->vertices[object->_indices[i*3+j]*3+2]);
      }
#if 0
      if(prev_material->flags & G3D_FLAG_FAC_TEXMAP)
      {
        glEnable(GL_TEXTURE_2D);
        glBindTexture( GL_TEXTURE_2D, evil);
        for(j=0; j<3; j++)
        {
          if(object->tex_nverts > object->_indices[i*3+j])
          {
#if DEBUG > 4
            g_printerr("tex: %f x %f\n",
              object->tex_vertices[object->_indices[i*3+j]*2+0],
              object->tex_vertices[object->_indices[i*3+j]*2+1]);
#endif
            glTexCoord2f(
              object->tex_vertices[object->_indices[i*3+j]*2+0],
              object->tex_vertices[object->_indices[i*3+j]*2+1]);
          }
#if DEBUG > 4
          else g_printerr("too few texture coords: %d (%d)\n",
                          object->_indices[i*3+j], object->tex_nverts);
#endif
        }
        glDisable(GL_TEXTURE_2D);
      }
#endif

      if(model->glflags & G3D_FLAG_GL_ALLTWOSIDE)
      {
        for(j=2; j>=0; j--)
        {
          if(object->_flags[i] & G3D_FLAG_FAC_NORMALS)
          {
            glNormal3f(object->_normals[(i*3+j)*3+0],
                       object->_normals[(i*3+j)*3+1],
                       object->_normals[(i*3+j)*3+2]);
          }
          else
          {
            glNormal3f(-object->_normals[(i*3+j)*3+0],
                       -object->_normals[(i*3+j)*3+1],
                       -object->_normals[(i*3+j)*3+2]);
          }
          glVertex3f(object->vertices[object->_indices[i*3+j]*3+0],
                     object->vertices[object->_indices[i*3+j]*3+1],
                     object->vertices[object->_indices[i*3+j]*3+2]);

          if(prev_material->flags & G3D_FLAG_FAC_TEXMAP)
          {
            if(object->tex_nverts > object->_indices[i*3+j])
            {
              glTexCoord2f(object->tex_vertices[object->_indices[i*3+j]*2+0],
                           object->tex_vertices[object->_indices[i*3+j]*2+1]);
            }
#if DEBUG > 4
            else g_printerr("too few texture coords: %d (%d)\n",
                            object->_indices[i*3+j], object->tex_nverts);
#endif
          }

        }
      }
    }
    glEnd();
    glDisable(GL_TEXTURE_2D);

/*****************************************************************************/
/* original drawing algorithm, _SHOULD_ be slower                            */

#else /* faces from GSList */

#define PX(i) (object->vertices[face->index_vertex[i]*3+0])
#define PY(i) (object->vertices[face->index_vertex[i]*3+1])
#define PZ(i) (object->vertices[face->index_vertex[i]*3+2])

    flist = object->faces;
    while(flist != NULL)
    {
      face = (G3DFace*)flist->data;

      g_return_if_fail(face != NULL);
      fcnt++;
#if 0
      g_printerr(" %d\n", fcnt);
#endif
      if(face->num_vertices < 3)
      {
        flist = flist->next;
        continue;
      }

      ax = PX(1) - PX(0);
      ay = PY(1) - PY(0);
      az = PZ(1) - PZ(0);

      bx = PX(face->num_vertices - 1) - PX(0);
      by = PY(face->num_vertices - 1) - PY(0);
      bz = PZ(face->num_vertices - 1) - PZ(0);

      nx = ay * bz - az * by;
      ny = az * bx - ax * bz;
      nz = ax * by - ay * bx;

      if(gl_normal_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
      {
        flist = flist->next;
        continue;
      }

      if(prev_index_cnt != face->num_vertices || prev_index_cnt > 4)
      {
        if(prev_index_cnt > 0) glEnd();
        prev_index_cnt = face->num_vertices;
        switch(face->num_vertices)
        {
          case 3:  glBegin(GL_TRIANGLES); break;
          case 4:  glBegin(GL_QUADS); break;
          default: glBegin(GL_POLYGON);
        }
      }

      if(prev_material != face->material)
      {
        GLfloat normspec[4] = { 0.0, 0.0, 0.0, 1.0 };

        g_return_if_fail(face->material != NULL);
        prev_material = face->material;
        glColor4f(face->material->r, face->material->g,
                  face->material->b, face->material->a);
        if(model->glflags & G3D_FLAG_GL_SPECULAR)
          glMaterialfv(GL_FRONT, GL_SPECULAR, face->material->specular);
        else glMaterialfv(GL_FRONT, GL_SPECULAR, normspec);
        if(model->glflags & G3D_FLAG_GL_SHININESS)
          glMaterialf(GL_FRONT, GL_SHININESS, face->material->shininess);
        else glMaterialf(GL_FRONT, GL_SHININESS, 0.0);
      }
#if DEBUG > 3
      g_printerr(" mat");
#endif
      if(nx != prev_nx || ny != prev_ny || nz != prev_nz)
      {
        prev_nx = nx;
        prev_ny = ny;
        prev_nz = nz;
        glNormal3f(nx, ny, nz);
      }
#if DEBUG > 3
      g_printerr(" nrm");
#endif
      for(i=0; i<face->num_vertices; i++)
      {
        if(face->flags & G3D_FLAG_FAC_NORMALS)
        {
          nx = face->normals[i*3+0];
          ny = face->normals[i*3+1];
          nz = face->normals[i*3+2];

          if(gl_normal_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
          {
            flist = flist->next;
            continue;
          }

          if(nx != prev_nx || ny != prev_ny || nz != prev_nz)
          {
            prev_nx = nx;
            prev_ny = ny;
            prev_nz = nz;
            glNormal3f(nx, ny, nz);
          }
        }
        glVertex3f(PX(i), PY(i), PZ(i));
      }

      if((face->material->flags & G3D_FLAG_MAT_TWOSIDE) ||
         (model->glflags & G3D_FLAG_GL_ALLTWOSIDE))
      {
        if(!(face->flags & G3D_FLAG_FAC_NORMALS))
        {
          prev_nx = nx = -nx;
          prev_ny = ny = -ny;
          prev_nz = nz = -nz;
          glNormal3f(nx, ny, nz);
        }
        for(i=face->num_vertices-1; i>=0; i--)
        {
          if(face->flags & G3D_FLAG_FAC_NORMALS)
          {
            nx = face->normals[i*3+0];
            ny = face->normals[i*3+1];
            nz = face->normals[i*3+2];

            if(gl_normal_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
            {
              flist = flist->next;
              continue;
            }

            if(nx != prev_nx || ny != prev_ny || nz != prev_nz)
            {
              prev_nx = nx;
              prev_ny = ny;
              prev_nz = nz;
              glNormal3f(nx, ny, nz);
            }
          }
          glVertex3f(PX(i), PY(i), PZ(i));
        }
      }
#if DEBUG > 3
      g_printerr(" fac");
#endif
      flist = flist->next;
    }
#endif /* draw faces from GSList instead of triangles */

#ifdef TIMING /* get time to draw one frame to compare algorithms */
    g_timer_stop(timer);

    if(avg_msec == 0)
    {
      gulong msec;
      gdouble sec;

      sec = g_timer_elapsed(timer, &msec);
      avg_msec = (gulong)sec * 1000000 + msec;
    }
    else
    {
      gulong msec, add;
      gdouble sec;

      sec = g_timer_elapsed(timer, &msec);
      add = (gulong)sec * 1000000 + msec;
      avg_msec = (avg_msec + add) / 2;
    }

    g_printerr("average time to render frame: %lu s\n", avg_msec);
#endif

#if DEBUG > 3
    g_printerr(" fcs");
    g_printerr(" }\n");
#endif

#ifdef OLD_ALGO
    if(prev_index_cnt > 0) glEnd();
    prev_index_cnt = -1;
    prev_nx = 0;
    prev_ny = 0;
    prev_nz = 0;
#endif

    olist = olist->next;
  } /* while olist != NULL */
#if DEBUG > 3
  g_printerr("gl.c: drawn...\n");
#endif
}

/*****************************************************************************/

int gl_normal_unify(GLfloat *nx, GLfloat *ny, GLfloat *nz)
{
  GLfloat r;

  r = sqrt(*nx * *nx + *ny * *ny + *nz * *nz);
  if(r < 0.000001) return EXIT_FAILURE;

  *nx /= r;
  *ny /= r;
  *nz /= r;
  return EXIT_SUCCESS;
}
